# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

import os

from pathlib import Path
from textwrap import dedent

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from .._builder import render_fragments, split_fragments
from .._writer import append_to_newsfile
from ..build import _main
from .helpers import read_pkg_resource, write


class WritingTests(TestCase):
    maxDiff = None

    def test_append_at_top(self):
        fragments = {
            "": {
                ("142", "misc", 0): "",
                ("1", "misc", 0): "",
                ("4", "feature", 0): "Stuff!",
                ("4", "feature", 1): "Second Stuff!",
                ("2", "feature", 0): "Foo added.",
                ("72", "feature", 0): "Foo added.",
            },
            "Names": {},
            "Web": {("3", "bugfix", 0): "Web fixed."},
        }

        definitions = {
            "feature": {"name": "Features", "showcontent": True},
            "bugfix": {"name": "Bugfixes", "showcontent": True},
            "misc": {"name": "Misc", "showcontent": False},
        }

        expected_output = """MyProject 1.0 (never)
=====================

Features
--------

- Foo added. (#2, #72)
- Stuff! (#4)
- Second Stuff! (#4)


Misc
----

- #1, #142


Names
-----

No significant changes.


Web
---

Bugfixes
~~~~~~~~

- Web fixed. (#3)


Old text.
"""

        tempdir = self.mktemp()
        os.makedirs(tempdir)

        with open(os.path.join(tempdir, "NEWS.rst"), "w") as f:
            f.write("Old text.\n")

        fragments = split_fragments(fragments, definitions)

        template = read_pkg_resource("templates/default.rst")

        append_to_newsfile(
            tempdir,
            "NEWS.rst",
            ".. towncrier release notes start\n",
            "",
            render_fragments(
                template,
                None,
                fragments,
                definitions,
                ["-", "~"],
                wrap=True,
                versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
            ),
            single_file=True,
        )

        with open(os.path.join(tempdir, "NEWS.rst")) as f:
            output = f.read()

        self.assertEqual(expected_output, output)

    def test_append_at_top_with_hint(self):
        """
        If there is a comment with C{.. towncrier release notes start},
        towncrier will add the version notes after it.
        """
        fragments = {
            "": {
                ("142", "misc", 0): "",
                ("1", "misc", 0): "",
                ("4", "feature", 0): "Stuff!",
                ("2", "feature", 0): "Foo added.",
                ("72", "feature", 0): "Foo added.",
                ("99", "feature", 0): "Foo! " * 100,
            },
            "Names": {},
            "Web": {("3", "bugfix", 0): "Web fixed."},
        }

        definitions = {
            "feature": {"name": "Features", "showcontent": True},
            "bugfix": {"name": "Bugfixes", "showcontent": True},
            "misc": {"name": "Misc", "showcontent": False},
        }

        expected_output = """Hello there! Here is some info.

.. towncrier release notes start

MyProject 1.0 (never)
=====================

Features
--------

- Foo added. (#2, #72)
- Stuff! (#4)
- Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo!
  Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo!
  Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo!
  Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo!
  Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo!
  Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo!
  Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! Foo! (#99)


Misc
----

- #1, #142


Names
-----

No significant changes.


Web
---

Bugfixes
~~~~~~~~

- Web fixed. (#3)


Old text.
"""

        tempdir = self.mktemp()
        write(
            os.path.join(tempdir, "NEWS.rst"),
            contents="""\
                Hello there! Here is some info.

                .. towncrier release notes start
                Old text.
            """,
            dedent=True,
        )

        fragments = split_fragments(fragments, definitions)

        template = read_pkg_resource("templates/default.rst")

        append_to_newsfile(
            tempdir,
            "NEWS.rst",
            ".. towncrier release notes start\n",
            "",
            render_fragments(
                template,
                None,
                fragments,
                definitions,
                ["-", "~"],
                wrap=True,
                versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
            ),
            single_file=True,
        )

        with open(os.path.join(tempdir, "NEWS.rst")) as f:
            output = f.read()

        self.assertEqual(expected_output, output)

    def test_multiple_file_no_start_string(self):
        """
        When no `start_string` is defined, the generated content is added at
        the start of the file.
        """
        tempdir = self.mktemp()
        os.makedirs(tempdir)

        definitions = {}
        fragments = split_fragments(fragments={}, definitions=definitions)

        template = read_pkg_resource("templates/default.rst")

        content = render_fragments(
            template=template,
            issue_format=None,
            fragments=fragments,
            definitions=definitions,
            underlines=["-", "~"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )

        append_to_newsfile(
            directory=tempdir,
            filename="NEWS.rst",
            start_string=None,
            top_line="",
            content=content,
            single_file=True,
        )

        with open(os.path.join(tempdir, "NEWS.rst")) as f:
            output = f.read()

        expected_output = dedent(
            """\
            MyProject 1.0 (never)
            =====================
        """
        )

        self.assertEqual(expected_output, output)

    def test_with_title_format_duplicate_version_raise(self):
        """
        When `single_file` enabled as default,
        and fragments of `version` already produced in the newsfile,
        a duplicate `build` will throw a ValueError.
        """
        runner = CliRunner()

        def do_build_once():
            with open("newsfragments/123.feature", "w") as f:
                f.write("Adds levitation")

            result = runner.invoke(
                _main,
                [
                    "--version",
                    "7.8.9",
                    "--name",
                    "foo",
                    "--date",
                    "01-01-2001",
                    "--yes",
                ],
            )
            return result

        # `single_file` default as true
        with runner.isolated_filesystem():
            with open("pyproject.toml", "w") as f:
                f.write(
                    dedent(
                        """
                    [tool.towncrier]
                    title_format="{name} {version} ({project_date})"
                    filename="{version}-notes.rst"
                    """
                    ).lstrip()
                )
            with open("{version}-notes.rst", "w") as f:
                f.write("Release Notes\n\n.. towncrier release notes start\n")
            os.mkdir("newsfragments")

            result = do_build_once()
            self.assertEqual(0, result.exit_code)
            # build again with the same version
            result = do_build_once()
            self.assertNotEqual(0, result.exit_code)
            self.assertIsInstance(result.exception, ValueError)
            self.assertSubstring(
                "already produced newsfiles for this version", result.exception.args[0]
            )

    def test_single_file_false_overwrite_duplicate_version(self):
        """
        When `single_file` disabled, multiple newsfiles generated and
        the content of which get overwritten each time.
        """
        runner = CliRunner()

        def do_build_once():
            with open("newsfragments/123.feature", "w") as f:
                f.write("Adds levitation")

            result = runner.invoke(
                _main,
                [
                    "--version",
                    "7.8.9",
                    "--name",
                    "foo",
                    "--date",
                    "01-01-2001",
                    "--yes",
                ],
            )
            return result

        # single_file = false
        with runner.isolated_filesystem():
            with open("pyproject.toml", "w") as f:
                f.write(
                    dedent(
                        """
                    [tool.towncrier]
                    single_file=false
                    title_format="{name} {version} ({project_date})"
                    filename="{version}-notes.rst"
                    """
                    ).lstrip()
                )
            os.mkdir("newsfragments")

            result = do_build_once()
            self.assertEqual(0, result.exit_code)
            # build again with the same version
            result = do_build_once()
            self.assertEqual(0, result.exit_code)

            notes = list(Path.cwd().glob("*-notes.rst"))
            self.assertEqual(1, len(notes))
            self.assertEqual("7.8.9-notes.rst", notes[0].name)

            with open(notes[0]) as f:
                output = f.read()

        expected_output = dedent(
            """\
            foo 7.8.9 (01-01-2001)
            ======================

            Features
            --------

            - Adds levitation (#123)
            """
        )

        self.assertEqual(expected_output, output)



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

"""
Responsible for getting the version and name from a project.
"""

from __future__ import annotations

import contextlib
import sys

from importlib import import_module
from importlib.metadata import version as metadata_version
from types import ModuleType


if sys.version_info >= (3, 10):
    from importlib.metadata import packages_distributions
else:
    from importlib_metadata import packages_distributions  # type: ignore


def _get_package(package_dir: str, package: str) -> ModuleType:
    try:
        module = import_module(package)
    except ImportError:
        # Package is not already available / installed.
        # Force importing it based on the source files.
        sys.path.insert(0, package_dir)

        try:
            module = import_module(package)
        except ImportError as e:
            err = f"tried to import {package}, but ran into this error: {e}"
            # NOTE: this might be redirected via "towncrier --draft > …".
            print(f"ERROR: {err}")
            raise
        finally:
            sys.path.pop(0)

    return module


def _get_metadata_version(package: str) -> str | None:
    """
    Try to get the version from the package metadata.
    """
    distributions = packages_distributions()
    distribution_names = distributions.get(package)
    if not distribution_names or len(distribution_names) != 1:
        # We can only determine the version if there is exactly one matching distribution.
        return None
    return metadata_version(distribution_names[0])


def get_version(package_dir: str, package: str) -> str:
    """
    Get the version of a package.

    Try to extract the version from the distribution version metadata that matches
    `package`, then fall back to looking for the package in `package_dir`.
    """
    version: str | None

    # First try to get the version from the package metadata.
    if version := _get_metadata_version(package):
        return version

    # When no version if found, fall back to looking for the package in `package_dir`.
    module = _get_package(package_dir, package)
    version = getattr(module, "__version__", None)
    if not version:
        raise Exception(
            f"No __version__ or metadata version info for the '{package}' package."
        )

    if isinstance(version, str):
        return version.strip()

    if isinstance(version, tuple):
        return ".".join(map(str, version)).strip()

    # Try duck-typing as an Incremental version.
    if hasattr(version, "base"):
        try:
            version = str(version.base()).strip()
            # Incremental uses `X.Y.rcN`.
            # Standardize on importlib (and PEP440) use of `X.YrcN`:
            return version.replace(".rc", "rc")  # type: ignore
        except TypeError:
            pass

    raise Exception(
        "Version must be a string, tuple, or an Incremental Version."
        " If you can't provide that, use the --version argument and specify one."
    )


def get_project_name(package_dir: str, package: str) -> str:
    module = _get_package(package_dir, package)
    version = getattr(module, "__version__", None)
    # Incremental has support for package names, try duck-typing it.
    with contextlib.suppress(AttributeError):
        return str(version.package)  # type: ignore

    return package.title()



# Copyright (c) Stephen Finucane, 2019
# See LICENSE for details.

"""
Create a new fragment.
"""

from __future__ import annotations

import os

from pathlib import Path
from typing import cast

import click

from ._builder import FragmentsPath
from ._settings import config_option_help, load_config_from_options


DEFAULT_CONTENT = "Add your info here"


@click.command(name="create")
@click.pass_context
@click.option(
    "--dir",
    "directory",
    default=None,
    metavar="PATH",
    help="Create fragment in directory. Default to current directory.",
)
@click.option(
    "--config",
    "config",
    default=None,
    metavar="FILE_PATH",
    help=config_option_help,
)
@click.option(
    "--edit/--no-edit",
    default=None,
    help="Open an editor for writing the newsfragment content.",
)
@click.option(
    "-c",
    "--content",
    type=str,
    default=DEFAULT_CONTENT,
    help="Sets the content of the new fragment.",
)
@click.option(
    "--section",
    type=str,
    help="The section to create the fragment for.",
)
@click.argument("filename", default="")
def _main(
    ctx: click.Context,
    directory: str | None,
    config: str | None,
    filename: str,
    edit: bool | None,
    content: str,
    section: str | None,
) -> None:
    """
    Create a new news fragment.

    If FILENAME is not provided, you'll be prompted to create it.

    Towncrier has a few standard types of news fragments, signified by the file
    extension.

    \b
    These are:
    * .feature - a new feature
    * .bugfix - a bug fix
    * .doc - a documentation improvement,
    * .removal - a deprecation or removal of public API,
    * .misc - an issue has been closed, but it is not of interest to users.

    If the FILENAME base is just '+' (to create a fragment not tied to an
    issue), it will be appended with a random hex string.
    """
    __main(ctx, directory, config, filename, edit, content, section)


def __main(
    ctx: click.Context,
    directory: str | None,
    config_path: str | None,
    filename: str,
    edit: bool | None,
    content: str,
    section: str | None,
) -> None:
    """
    The main entry point.
    """
    base_directory, config = load_config_from_options(directory, config_path)

    filename_ext = ""
    if config.create_add_extension:
        ext = os.path.splitext(config.filename)[1]
        if ext.lower() in (".rst", ".md"):
            filename_ext = ext

    section_provided = section is not None
    if not section_provided:
        # Get the default section.
        if len(config.sections) == 1:
            section = next(iter(config.sections))
        else:
            # If there are multiple sections then the first without a path is the default
            # section, otherwise it's the first defined section.
            for (
                section_name,
                section_dir,
            ) in config.sections.items():  # pragma: no branch
                if not section_dir:
                    section = section_name
                    break
            if section is None:
                section = list(config.sections.keys())[0]

    if section not in config.sections:
        # Raise a click exception with the correct parameter.
        section_param = None
        for p in ctx.command.params:  # pragma: no branch
            if p.name == "section":
                section_param = p
                break
        expected_sections = ", ".join(f"'{s}'" for s in config.sections)
        raise click.BadParameter(
            f"expected one of {expected_sections}",
            param=section_param,
        )
    section = cast(str, section)

    if not filename:
        if not section_provided:
            sections = list(config.sections)
            if len(sections) > 1:
                click.echo("Pick a section:")
                default_section_index = None
                for i, s in enumerate(sections):
                    click.echo(f" {i+1}: {s or '(primary)'}")
                    if not default_section_index and s == section:
                        default_section_index = str(i + 1)
                section_index = click.prompt(
                    "Section",
                    type=click.Choice([str(i + 1) for i in range(len(sections))]),
                    default=default_section_index,
                )
                section = sections[int(section_index) - 1]
        prompt = "Issue number"
        # Add info about adding orphan if config is set.
        if config.orphan_prefix:
            prompt += f" (`{config.orphan_prefix}` if none)"
        issue = click.prompt(prompt)
        fragment_type = click.prompt(
            "Fragment type",
            type=click.Choice(list(config.types)),
        )
        filename = f"{issue}.{fragment_type}"
        if edit is None and content == DEFAULT_CONTENT:
            edit = True

    file_dir, file_basename = os.path.split(filename)
    if config.orphan_prefix and file_basename.startswith(f"{config.orphan_prefix}."):
        # Append a random hex string to the orphan news fragment base name.
        filename = os.path.join(
            file_dir,
            (
                f"{config.orphan_prefix}{os.urandom(4).hex()}"
                f"{file_basename[len(config.orphan_prefix):]}"
            ),
        )
    filename_parts = filename.split(".")
    if len(filename_parts) < 2 or (
        filename_parts[-1] not in config.types
        and filename_parts[-2] not in config.types
    ):
        raise click.BadParameter(
            "Expected filename '{}' to be of format '{{name}}.{{type}}', "
            "where '{{name}}' is an arbitrary slug and '{{type}}' is "
            "one of: {}".format(filename, ", ".join(config.types))
        )
    if filename_parts[-1] in config.types and filename_ext:
        filename += filename_ext

    get_fragments_path = FragmentsPath(base_directory, config)
    fragments_directory = get_fragments_path(section_directory=config.sections[section])

    if not os.path.exists(fragments_directory):
        os.makedirs(fragments_directory)

    segment_file = os.path.join(fragments_directory, filename)

    retry = 0
    if filename.split(".")[-1] not in config.types:
        filename, extra_ext = os.path.splitext(filename)
    else:
        extra_ext = ""
    while os.path.exists(segment_file):
        retry += 1
        segment_file = os.path.join(
            fragments_directory, f"{filename}.{retry}{extra_ext}"
        )

    if edit:
        if content == DEFAULT_CONTENT:
            content = ""
        content = _get_news_content_from_user(content, extension=filename_ext)
        if not content:
            click.echo("Aborted creating news fragment due to empty message.")
            ctx.exit(1)

    add_newline = bool(
        config.create_eof_newline and content and not content.endswith("\n")
    )
    Path(segment_file).write_text(content + "\n" * add_newline, encoding="utf-8")

    click.echo(f"Created news fragment at {segment_file}")


def _get_news_content_from_user(message: str, extension: str = "") -> str:
    initial_content = """
# Please write your news content. Lines starting with '#' will be ignored, and
# an empty message aborts.
"""
    if message:
        initial_content = f"{message}\n{initial_content}"
    content = click.edit(initial_content, extension=extension or ".txt")
    if content is None:
        return message
    all_lines = content.split("\n")
    lines = [line.rstrip() for line in all_lines if not line.lstrip().startswith("#")]
    return "\n".join(lines).strip()


if __name__ == "__main__":  # pragma: no cover
    _main()



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

import os
import sys

from importlib.metadata import version as metadata_version

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from .._project import get_project_name, get_version
from .._shell import cli as towncrier_cli
from .helpers import write


towncrier_cli.name = "towncrier"


class VersionFetchingTests(TestCase):
    def test_str(self):
        """
        A str __version__ will be picked up.
        """
        temp = self.mktemp()
        os.makedirs(os.path.join(temp, "mytestproj"))

        with open(os.path.join(temp, "mytestproj", "__init__.py"), "w") as f:
            f.write("__version__ = '1.2.3'")

        version = get_version(temp, "mytestproj")
        self.assertEqual(version, "1.2.3")

    def test_tuple(self):
        """
        A tuple __version__ will be picked up.
        """
        temp = self.mktemp()
        os.makedirs(os.path.join(temp, "mytestproja"))

        with open(os.path.join(temp, "mytestproja", "__init__.py"), "w") as f:
            f.write("__version__ = (1, 3, 12)")

        version = get_version(temp, "mytestproja")
        self.assertEqual(version, "1.3.12")

    def test_incremental(self):
        """
        An incremental-like Version __version__  is picked up.
        """
        temp = self.mktemp()
        os.makedirs(temp)
        os.makedirs(os.path.join(temp, "mytestprojinc"))

        with open(os.path.join(temp, "mytestprojinc", "__init__.py"), "w") as f:
            f.write(
                """
class Version:
    '''
    This is emulating a Version object from incremental.
    '''

    def __init__(self, *version_parts):
        self.version = version_parts
        self.package = "mytestprojinc"

    def base(self):
        return '.'.join(map(str, self.version))

__version__ = Version(1, 3, 12, "rc1")
                """
            )

        version = get_version(temp, "mytestprojinc")
        self.assertEqual(version, "1.3.12rc1")

        project = get_project_name(temp, "mytestprojinc")
        self.assertEqual(project, "mytestprojinc")

    def test_not_incremental(self):
        """
        An exception is raised when the version could not be detected.
        For this test we use an incremental-like object,
        that has the `base` method, but that method
        does not match the return type for `incremental`.
        """
        temp = self.mktemp()
        os.makedirs(os.path.join(temp, "mytestprojnotinc"))

        with open(os.path.join(temp, "mytestprojnotinc", "__init__.py"), "w") as f:
            f.write(
                """
class WeirdVersion:
    def base(self, some_arg):
        return "shouldn't get here"


__version__ = WeirdVersion()
"""
            )
        with self.assertRaises(Exception) as e:
            get_version(temp, "mytestprojnotinc")

        self.assertEqual(
            (
                "Version must be a string, tuple, or an Incremental Version. "
                "If you can't provide that, use the --version argument and "
                "specify one.",
            ),
            e.exception.args,
        )

    def test_version_from_metadata(self):
        """
        A version from package metadata is picked up.
        """
        version = get_version(".", "towncrier")
        self.assertEqual(metadata_version("towncrier"), version)

    def _setup_missing(self):
        """
        Create a minimalistic project with missing metadata in a temporary
        directory.
        """
        tmp_dir = self.mktemp()
        pkg = os.path.join(tmp_dir, "missing")
        os.makedirs(pkg)
        init = os.path.join(tmp_dir, "__init__.py")

        write(init, "# nope\n")

        return tmp_dir

    def test_missing_version(self):
        """
        Missing __version__ string leads to an exception.
        """
        tmp_dir = self._setup_missing()

        with self.assertRaises(Exception) as e:
            # The 'missing' package has no __version__ string.
            get_version(tmp_dir, "missing")

        self.assertEqual(
            ("No __version__ or metadata version info for the 'missing' package.",),
            e.exception.args,
        )

    def test_missing_version_project_name(self):
        """
        Missing __version__ string leads to the package name becoming the
        project name.
        """
        tmp_dir = self._setup_missing()

        self.assertEqual("Missing", get_project_name(tmp_dir, "missing"))

    def test_unknown_type(self):
        """
        A __version__ of unknown type will lead to an exception.
        """
        temp = self.mktemp()
        os.makedirs(os.path.join(temp, "mytestprojb"))

        with open(os.path.join(temp, "mytestprojb", "__init__.py"), "w") as f:
            f.write("__version__ = object()")

        self.assertRaises(Exception, get_version, temp, "mytestprojb")

        self.assertEqual("Mytestprojb", get_project_name(temp, "mytestprojb"))

    def test_import_fails(self):
        """
        An exception is raised when getting the version failed due to missing Python package files.
        """
        with self.assertRaises(ModuleNotFoundError):
            get_version(".", "projectname_without_any_files")

    def test_already_installed_import(self):
        """
        An already installed package will be checked before cwd-found packages.
        """
        project_name = "mytestproj_already_installed_import"

        temp = self.mktemp()
        os.makedirs(os.path.join(temp, project_name))

        with open(os.path.join(temp, project_name, "__init__.py"), "w") as f:
            f.write("__version__ = (1, 3, 12)")

        sys_path_temp = self.mktemp()
        os.makedirs(os.path.join(sys_path_temp, project_name))

        with open(os.path.join(sys_path_temp, project_name, "__init__.py"), "w") as f:
            f.write("__version__ = (2, 1, 5)")

        sys.path.insert(0, sys_path_temp)
        self.addCleanup(sys.path.pop, 0)

        version = get_version(temp, project_name)

        self.assertEqual(version, "2.1.5")

    def test_installed_package_found_when_no_source_present(self):
        """
        The version from the installed package is returned when there is no
        package present at the provided source directory.
        """
        project_name = "mytestproj_only_installed"

        sys_path_temp = self.mktemp()
        os.makedirs(os.path.join(sys_path_temp, project_name))

        with open(os.path.join(sys_path_temp, project_name, "__init__.py"), "w") as f:
            f.write("__version__ = (3, 14)")

        sys.path.insert(0, sys_path_temp)
        self.addCleanup(sys.path.pop, 0)

        version = get_version("some non-existent directory", project_name)

        self.assertEqual(version, "3.14")


class InvocationTests(TestCase):
    def test_dash_m(self):
        """
        `python -m towncrier` invokes the main entrypoint.
        """
        runner = CliRunner()
        temp = self.mktemp()
        new_dir = os.path.join(temp, "dashm")
        os.makedirs(new_dir)
        orig_dir = os.getcwd()
        try:
            os.chdir(new_dir)
            with open("pyproject.toml", "w") as f:
                f.write("[tool.towncrier]\n" 'directory = "news"\n')
            os.makedirs("news")
            result = runner.invoke(towncrier_cli, ["--help"])
            self.assertIn("[OPTIONS] COMMAND [ARGS]...", result.stdout)
            self.assertRegex(result.stdout, r".*--help\s+Show this message and exit.")
        finally:
            os.chdir(orig_dir)

    def test_version(self):
        """
        `--version` command line option is available to show the current production version.
        """
        runner = CliRunner()
        result = runner.invoke(towncrier_cli, ["--version"])
        self.assertTrue(result.output.startswith("towncrier, version 2"))



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

from __future__ import annotations

import os

from subprocess import STDOUT, CalledProcessError, call, check_output


def remove_files(fragment_filenames: list[str]) -> None:
    if not fragment_filenames:
        return

    # Filter out files that are unknown to git
    try:
        git_fragments = check_output(
            ["git", "ls-files"] + fragment_filenames, encoding="utf-8"
        ).split("\n")
    except CalledProcessError:
        # we may not be in a git repository
        git_fragments = []

    git_fragments = [os.path.abspath(f) for f in git_fragments if os.path.isfile(f)]
    call(["git", "rm", "--quiet", "--force"] + git_fragments)
    unknown_fragments = set(fragment_filenames) - set(git_fragments)
    for unknown_fragment in unknown_fragments:
        os.remove(unknown_fragment)


def stage_newsfile(directory: str, filename: str) -> None:
    call(["git", "add", os.path.join(directory, filename)])


def get_remote_branches(base_directory: str) -> list[str]:
    output = check_output(
        ["git", "branch", "-r"], cwd=base_directory, encoding="utf-8", stderr=STDOUT
    )

    return [branch.strip() for branch in output.strip().splitlines()]


def list_changed_files_compared_to_branch(
    base_directory: str, compare_with: str
) -> list[str]:
    output = check_output(
        ["git", "diff", "--name-only", compare_with + "..."],
        cwd=base_directory,
        encoding="utf-8",
        stderr=STDOUT,
    )

    return output.strip().splitlines()



# Copyright (c) Povilas Kanapickas, 2019
# See LICENSE for details.

from textwrap import dedent

from twisted.trial.unittest import TestCase

from .._builder import parse_newfragment_basename, render_fragments


class TestParseNewsfragmentBasename(TestCase):
    def test_simple(self):
        """<number>.<category> generates a counter value of 0."""
        self.assertEqual(
            parse_newfragment_basename("123.feature", ["feature"]),
            ("123", "feature", 0),
        )

    def test_invalid_category(self):
        """Files without a valid category are rejected."""
        self.assertEqual(
            parse_newfragment_basename("README.ext", ["feature"]),
            (None, None, None),
        )
        self.assertEqual(
            parse_newfragment_basename("README", ["feature"]),
            (None, None, None),
        )

    def test_counter(self):
        """<number>.<category>.<counter> generates a custom counter value."""
        self.assertEqual(
            parse_newfragment_basename("123.feature.1", ["feature"]),
            ("123", "feature", 1),
        )

    def test_counter_with_extension(self):
        """File extensions are ignored."""
        self.assertEqual(
            parse_newfragment_basename("123.feature.1.ext", ["feature"]),
            ("123", "feature", 1),
        )

    def test_ignores_extension(self):
        """File extensions are ignored."""
        self.assertEqual(
            parse_newfragment_basename("123.feature.ext", ["feature"]),
            ("123", "feature", 0),
        )

    def test_non_numeric_issue(self):
        """Non-numeric issue identifiers are preserved verbatim."""
        self.assertEqual(
            parse_newfragment_basename("baz.feature", ["feature"]),
            ("baz", "feature", 0),
        )

    def test_non_numeric_issue_with_extension(self):
        """File extensions are ignored."""
        self.assertEqual(
            parse_newfragment_basename("baz.feature.ext", ["feature"]),
            ("baz", "feature", 0),
        )

    def test_dots_in_issue_name(self):
        """Non-numeric issue identifiers are preserved verbatim."""
        self.assertEqual(
            parse_newfragment_basename("baz.1.2.feature", ["feature"]),
            ("baz.1.2", "feature", 0),
        )

    def test_dots_in_issue_name_invalid_category(self):
        """Files without a valid category are rejected."""
        self.assertEqual(
            parse_newfragment_basename("baz.1.2.notfeature", ["feature"]),
            (None, None, None),
        )

    def test_dots_in_issue_name_and_counter(self):
        """Non-numeric issue identifiers are preserved verbatim."""
        self.assertEqual(
            parse_newfragment_basename("baz.1.2.feature.3", ["feature"]),
            ("baz.1.2", "feature", 3),
        )

    def test_strip(self):
        """Leading spaces and subsequent leading zeros are stripped
        when parsing newsfragment names into issue numbers etc.
        """
        self.assertEqual(
            parse_newfragment_basename("  007.feature", ["feature"]),
            ("7", "feature", 0),
        )

    def test_strip_with_counter(self):
        """Leading spaces and subsequent leading zeros are stripped
        when parsing newsfragment names into issue numbers etc.
        """
        self.assertEqual(
            parse_newfragment_basename("  007.feature.3", ["feature"]),
            ("7", "feature", 3),
        )

    def test_orphan(self):
        """Orphaned snippets must remain the orphan marker in the issue
        identifier."""
        self.assertEqual(
            parse_newfragment_basename("+orphan.feature", ["feature"]),
            ("+orphan", "feature", 0),
        )

    def test_orphan_with_number(self):
        """Orphaned snippets can contain numbers in the identifier."""
        self.assertEqual(
            parse_newfragment_basename("+123_orphan.feature", ["feature"]),
            ("+123_orphan", "feature", 0),
        )
        self.assertEqual(
            parse_newfragment_basename("+orphan_123.feature", ["feature"]),
            ("+orphan_123", "feature", 0),
        )

    def test_orphan_with_dotted_number(self):
        """Orphaned snippets can contain numbers with dots in the
        identifier."""
        self.assertEqual(
            parse_newfragment_basename("+12.3_orphan.feature", ["feature"]),
            ("+12.3_orphan", "feature", 0),
        )
        self.assertEqual(
            parse_newfragment_basename("+orphan_12.3.feature", ["feature"]),
            ("+orphan_12.3", "feature", 0),
        )

    def test_orphan_all_digits(self):
        """Orphaned snippets can consist of only digits."""
        self.assertEqual(
            parse_newfragment_basename("+123.feature", ["feature"]),
            ("+123", "feature", 0),
        )


class TestNewsFragmentsOrdering(TestCase):
    """
    Tests to ensure that issues are ordered correctly in the output.

    This tests both ordering of issues within a fragment and ordering of
    fragments within a section.
    """

    template = dedent(
        """
    {% for section_name, category in sections.items() %}
    {% if section_name %}# {{ section_name }}{% endif %}
    {%- for category_name, issues in category.items() %}
    ## {{ category_name }}
    {% for issue, numbers in issues.items() %}
    - {{ issue }}{% if numbers %} ({{ numbers|join(', ') }}){% endif %}

    {% endfor %}
    {% endfor -%}
    {% endfor -%}
    """
    )

    def render(self, fragments):
        return render_fragments(
            template=self.template,
            issue_format=None,
            fragments=fragments,
            definitions={},
            underlines=[],
            wrap=False,
            versiondata={},
        )

    def test_ordering(self):
        """
        Issues are ordered first by the non-text component, then by their number.

        For backwards compatibility, issues with no number are grouped first and issues
        which are only a number are grouped last.

        Orphan news fragments are always last, sorted by their text.
        """
        output = self.render(
            {
                "": {
                    "feature": {
                        "Added Cheese": ["10", "gh-25", "gh-3", "4"],
                        "Added Fish": [],
                        "Added Bread": [],
                        "Added Milk": ["gh-1"],
                        "Added Eggs": ["gh-2", "random"],
                    }
                }
            },
        )
        # "Eggs" are first because they have an issue with no number, and the first
        # issue for each fragment is what is used for sorting the overall list.
        assert output == dedent(
            """
            ## feature
            - Added Eggs (random, gh-2)
            - Added Milk (gh-1)
            - Added Cheese (gh-3, gh-25, #4, #10)
            - Added Bread
            - Added Fish
"""
        )



# SPDX-FileCopyrightText: 2015 Heungsub Lee <19982+sublee@users.noreply.github.com>
#
# SPDX-License-Identifier: BSD-3-Clause

# Vendored from
# https://github.com/click-contrib/click-default-group/tree/b671ae5325d186fe5ea7abb584f15852a1e931aa
# Because the PyPI package could not be installed on modern Pips anymore and
# the project looks unmaintaintained.

"""
   click_default_group
   ~~~~~~~~~~~~~~~~~~~

   Define a default subcommand by `default=True`:

   .. sourcecode:: python

      import click
      from click_default_group import DefaultGroup

      @click.group(cls=DefaultGroup, default_if_no_args=True)
      def cli():
          pass

      @cli.command(default=True)
      def foo():
          click.echo('foo')

      @cli.command()
      def bar():
          click.echo('bar')

   Then you can invoke that without explicit subcommand name:

   .. sourcecode:: console

      $ cli.py --help
      Usage: cli.py [OPTIONS] COMMAND [ARGS]...

      Options:
        --help    Show this message and exit.

      Command:
        foo*
        bar

      $ cli.py
      foo
      $ cli.py foo
      foo
      $ cli.py bar
      bar

"""
import warnings

import click


__all__ = ["DefaultGroup"]
__version__ = "1.2.2"


class DefaultGroup(click.Group):
    """Invokes a subcommand marked with `default=True` if any subcommand not
    chosen.

    :param default_if_no_args: resolves to the default command if no arguments
                               passed.

    """

    def __init__(self, *args, **kwargs):
        # To resolve as the default command.
        if not kwargs.get("ignore_unknown_options", True):
            raise ValueError("Default group accepts unknown options")
        self.ignore_unknown_options = True
        self.default_cmd_name = kwargs.pop("default", None)
        self.default_if_no_args = kwargs.pop("default_if_no_args", False)
        super().__init__(*args, **kwargs)

    def set_default_command(self, command):
        """Sets a command function as the default command."""
        cmd_name = command.name
        self.add_command(command)
        self.default_cmd_name = cmd_name

    def parse_args(self, ctx, args):
        if not args and self.default_if_no_args:
            args.insert(0, self.default_cmd_name)
        return super().parse_args(ctx, args)

    def get_command(self, ctx, cmd_name):
        if cmd_name not in self.commands:
            # No command name matched.
            ctx.arg0 = cmd_name
            cmd_name = self.default_cmd_name
        return super().get_command(ctx, cmd_name)

    def resolve_command(self, ctx, args):
        base = super()
        cmd_name, cmd, args = base.resolve_command(ctx, args)
        if hasattr(ctx, "arg0"):
            args.insert(0, ctx.arg0)
            cmd_name = cmd.name
        return cmd_name, cmd, args

    def format_commands(self, ctx, formatter):
        formatter = DefaultCommandFormatter(self, formatter, mark="*")
        return super().format_commands(ctx, formatter)

    def command(self, *args, **kwargs):
        default = kwargs.pop("default", False)
        decorator = super().command(*args, **kwargs)
        if not default:
            return decorator
        warnings.warn(
            "Use default param of DefaultGroup or " "set_default_command() instead",
            DeprecationWarning,
        )

        def _decorator(f):
            cmd = decorator(f)
            self.set_default_command(cmd)
            return cmd

        return _decorator


class DefaultCommandFormatter:
    """Wraps a formatter to mark a default command."""

    def __init__(self, group, formatter, mark="*"):
        self.group = group
        self.formatter = formatter
        self.mark = mark

    def __getattr__(self, attr):
        return getattr(self.formatter, attr)

    def write_dl(self, rows, *args, **kwargs):
        rows_ = []
        for cmd_name, help in rows:
            if cmd_name == self.group.default_cmd_name:
                rows_.insert(0, (cmd_name + self.mark, help))
            else:
                rows_.append((cmd_name, help))
        return self.formatter.write_dl(rows_, *args, **kwargs)



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

"""
towncrier, a builder for your news files.
"""

from __future__ import annotations


__all__ = ["__version__"]


def __getattr__(name: str) -> str:
    if name != "__version__":
        raise AttributeError(f"module {__name__} has no attribute {name}")

    import warnings

    from ._version import __version__

    warnings.warn(
        "Accessing towncrier.__version__ is deprecated and will be "
        "removed in a future release. Use importlib.metadata directly "
        "to query for towncrier's packaging metadata.",
        DeprecationWarning,
        stacklevel=2,
    )

    return __version__



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

import os
import string

from pathlib import Path
from textwrap import dedent
from unittest import mock

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from ..create import DEFAULT_CONTENT, _main
from .helpers import setup_simple_project, with_isolated_runner


class TestCli(TestCase):
    maxDiff = None

    def _test_success(
        self,
        content=None,
        config=None,
        mkdir=True,
        additional_args=None,
        eof_newline=True,
    ):
        runner = CliRunner()

        with runner.isolated_filesystem():
            setup_simple_project(config=config, mkdir_newsfragments=mkdir)

            args = ["123.feature.rst"]
            if content is None:
                content = [DEFAULT_CONTENT]
            if additional_args is not None:
                args.extend(additional_args)
            result = runner.invoke(_main, args)

            self.assertEqual(["123.feature.rst"], os.listdir("foo/newsfragments"))

            if eof_newline:
                content.append("")
            with open("foo/newsfragments/123.feature.rst") as fh:
                self.assertEqual("\n".join(content), fh.read())

        self.assertEqual(0, result.exit_code)

    def test_basics(self):
        """Ensure file created where output directory already exists."""
        self._test_success(mkdir=True)

    def test_directory_created(self):
        """Ensure both file and output directory created if necessary."""
        self._test_success(mkdir=False)

    def test_edit_without_comments(self):
        """Create file with dynamic content."""
        content = ["This is line 1", "This is line 2"]
        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "\n".join(content)
            self._test_success(content=content, additional_args=["--edit"])
            mock_edit.assert_called_once_with(
                "\n# Please write your news content. Lines starting "
                "with '#' will be ignored, and\n# an empty message aborts.\n",
                extension=".rst",
            )

    def test_edit_with_comment(self):
        """Create file editly with ignored line."""
        content = ["This is line 1", "This is line 2"]
        comment = "# I am ignored"
        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "\n".join(content[:1] + [comment] + content[1:])
            self._test_success(content=content, additional_args=["--edit"])

    def test_edit_abort(self):
        """Create file editly and abort."""
        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = None

            runner = CliRunner()

            with runner.isolated_filesystem():
                setup_simple_project(config=None, mkdir_newsfragments=True)
                result = runner.invoke(_main, ["123.feature.rst", "--edit"])
                self.assertEqual([], os.listdir("foo/newsfragments"))
                self.assertEqual(1, result.exit_code)

    def test_edit_markdown_extension(self):
        """
        The temporary file extension used when editing is ``.md`` if the main filename
        also uses that extension.
        """

        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "This is line 1"
            self._test_success(
                content=["This is line 1"],
                config=dedent(
                    """\
                    [tool.towncrier]
                    package = "foo"
                    filename = "README.md"
                    """
                ),
                additional_args=["--edit"],
            )
            mock_edit.assert_called_once_with(
                "\n# Please write your news content. Lines starting "
                "with '#' will be ignored, and\n# an empty message aborts.\n",
                extension=".md",
            )

    def test_edit_unknown_extension(self):
        """
        The temporary file extension used when editing is ``.txt`` if it the main
        filename isn't ``.rst`` or ``.md``.
        """

        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "This is line 1"
            self._test_success(
                content=["This is line 1"],
                config=dedent(
                    """\
                    [tool.towncrier]
                    package = "foo"
                    filename = "README.FIRST"
                    """
                ),
                additional_args=["--edit"],
            )
            mock_edit.assert_called_once_with(
                "\n# Please write your news content. Lines starting "
                "with '#' will be ignored, and\n# an empty message aborts.\n",
                extension=".txt",
            )

    def test_content(self):
        """
        When creating a new fragment the content can be passed as a
        command line argument.
        The text editor is not invoked.
        """
        content_line = "This is a content"
        self._test_success(content=[content_line], additional_args=["-c", content_line])

    def test_content_without_eof_newline(self):
        """
        When creating a new fragment the content can be passed as a command line
        argument. The text editor is not invoked, and no eof newline is added if the
        config option is set.
        """
        config = dedent(
            """\
            [tool.towncrier]
            package = "foo"
            create_eof_newline = false
            """
        )
        content_line = "This is a content"
        self._test_success(
            content=[content_line],
            additional_args=["-c", content_line],
            config=config,
            eof_newline=False,
        )

    def test_message_and_edit(self):
        """
        When creating a new message, a initial content can be passed via
        the command line and continue modifying the content by invoking the
        text editor.
        """
        content_line = "This is a content line"
        edit_content = ["This is line 1", "This is line 2"]
        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "\n".join(edit_content)
            self._test_success(
                content=edit_content, additional_args=["-c", content_line, "--edit"]
            )
            mock_edit.assert_called_once_with(
                f"{content_line}\n\n# Please write your news content. Lines starting "
                "with '#' will be ignored, and\n# an empty message aborts.\n",
                extension=".rst",
            )

    def test_different_directory(self):
        """Ensure non-standard directories are used."""
        runner = CliRunner()
        config = dedent(
            """\
            [tool.towncrier]
            directory = "releasenotes"
            """
        )

        with runner.isolated_filesystem():
            setup_simple_project(config=config, mkdir_newsfragments=False)
            os.mkdir("releasenotes")

            result = runner.invoke(_main, ["123.feature.rst"])

            self.assertEqual(["123.feature.rst"], os.listdir("releasenotes"))

        self.assertEqual(0, result.exit_code)

    def test_invalid_section(self):
        """Ensure creating a path without a valid section is rejected."""
        runner = CliRunner()

        with runner.isolated_filesystem():
            setup_simple_project()

            self.assertEqual([], os.listdir("foo/newsfragments"))

            result = runner.invoke(_main, ["123.foobar.rst"])

            self.assertEqual([], os.listdir("foo/newsfragments"))

        self.assertEqual(type(result.exception), SystemExit, result.exception)
        self.assertIn(
            "Expected filename '123.foobar.rst' to be of format", result.output
        )

    @with_isolated_runner
    def test_custom_extension(self, runner: CliRunner):
        """Ensure we can still create fragments with custom extensions."""
        setup_simple_project()
        frag_path = Path("foo", "newsfragments")

        result = runner.invoke(_main, ["123.feature.txt"])
        self.assertEqual(result.exit_code, 0, result.output)

        fragments = [f.name for f in frag_path.iterdir()]
        # No '.rst' extension added.
        self.assertEqual(fragments, ["123.feature.txt"])

    @with_isolated_runner
    def test_md_filename_extension(self, runner: CliRunner):
        """Ensure changelog filename extension is used if .md"""
        setup_simple_project(extra_config='filename = "changes.md"')
        frag_path = Path("foo", "newsfragments")

        result = runner.invoke(_main, ["123.feature"])
        self.assertEqual(result.exit_code, 0, result.output)

        fragments = [f.name for f in frag_path.iterdir()]
        # No '.rst' extension added.
        self.assertEqual(fragments, ["123.feature.md"])

    @with_isolated_runner
    def test_no_filename_extension(self, runner: CliRunner):
        """
        When the NEWS filename has no extension, new fragments are will not have an
        extension added.
        """
        # The name of the file where towncrier will generate
        # the final release notes is named `RELEASE_NOTES`
        # for this test (with no file extension).
        setup_simple_project(extra_config='filename = "RELEASE_NOTES"')
        frag_path = Path("foo", "newsfragments")

        result = runner.invoke(_main, ["123.feature"])
        self.assertEqual(result.exit_code, 0, result.output)

        fragments = [f.name for f in frag_path.iterdir()]
        # No '.rst' extension added.
        self.assertEqual(fragments, ["123.feature"])

    @with_isolated_runner
    def test_file_exists(self, runner: CliRunner):
        """Ensure we don't overwrite existing files."""
        setup_simple_project()
        frag_path = Path("foo", "newsfragments")

        for _ in range(3):
            result = runner.invoke(_main, ["123.feature"])
            self.assertEqual(result.exit_code, 0, result.output)

        fragments = [f.name for f in frag_path.iterdir()]
        self.assertEqual(
            sorted(fragments),
            [
                "123.feature.1.rst",
                "123.feature.2.rst",
                "123.feature.rst",
            ],
        )

    @with_isolated_runner
    def test_file_exists_no_ext(self, runner: CliRunner):
        """
        Ensure we don't overwrite existing files with when not adding filename
        extensions.
        """

        setup_simple_project(extra_config="create_add_extension = false")
        frag_path = Path("foo", "newsfragments")

        for _ in range(3):
            result = runner.invoke(_main, ["123.feature"])
            self.assertEqual(result.exit_code, 0, result.output)

        fragments = [f.name for f in frag_path.iterdir()]
        self.assertEqual(
            sorted(fragments),
            [
                "123.feature",
                "123.feature.1",
                "123.feature.2",
            ],
        )

    @with_isolated_runner
    def test_file_exists_with_ext(self, runner: CliRunner):
        """
        Ensure we don't overwrite existing files when using an extension after the
        fragment type.
        """
        setup_simple_project()
        frag_path = Path("foo", "newsfragments")

        for _ in range(3):
            result = runner.invoke(_main, ["123.feature.rst"])
            self.assertEqual(result.exit_code, 0, result.output)

        fragments = [f.name for f in frag_path.iterdir()]
        self.assertEqual(
            sorted(fragments),
            [
                "123.feature.1.rst",
                "123.feature.2.rst",
                "123.feature.rst",
            ],
        )

    @with_isolated_runner
    def test_without_filename(self, runner: CliRunner):
        """
        When no filename is provided, the user is prompted for one.
        """
        setup_simple_project()

        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "Edited content"
            result = runner.invoke(_main, input="123\nfeature\n")
            self.assertFalse(result.exception, result.output)
            mock_edit.assert_called_once()
        expected = os.path.join(os.getcwd(), "foo", "newsfragments", "123.feature.rst")
        self.assertEqual(
            result.output,
            f"""Issue number (`+` if none): 123
Fragment type (feature, bugfix, doc, removal, misc): feature
Created news fragment at {expected}
""",
        )
        with open(expected) as f:
            self.assertEqual(f.read(), "Edited content\n")

    @with_isolated_runner
    def test_without_filename_orphan(self, runner: CliRunner):
        """
        The user can create an orphan fragment from the interactive prompt.
        """
        setup_simple_project()

        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "Orphan content"
            result = runner.invoke(_main, input="+\nfeature\n")
            self.assertFalse(result.exception, result.output)
            mock_edit.assert_called_once()
        expected = os.path.join(os.getcwd(), "foo", "newsfragments", "+")
        self.assertTrue(
            result.output.startswith(
                f"""Issue number (`+` if none): +
Fragment type (feature, bugfix, doc, removal, misc): feature
Created news fragment at {expected}"""
            ),
            result.output,
        )
        # Check that the file was created with a random name
        created_line = result.output.strip().rsplit("\n", 1)[-1]
        # Get file names in the newsfragments directory.
        files = os.listdir(os.path.join(os.getcwd(), "foo", "newsfragments"))
        # Check that the file name is in the created line.
        created_fragment = created_line.split(" ")[-1]
        self.assertIn(Path(created_fragment).name, files)
        with open(created_fragment) as f:
            self.assertEqual(f.read(), "Orphan content\n")

    @with_isolated_runner
    def test_without_filename_no_orphan_config(self, runner: CliRunner):
        """
        If an empty orphan prefix is set, orphan creation is turned off from interactive
        prompt.
        """
        setup_simple_project(extra_config='orphan_prefix = ""')

        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "Edited content"
            result = runner.invoke(_main, input="+\nfeature\n")
            self.assertFalse(result.exception, result.output)
            mock_edit.assert_called_once()
        expected = os.path.join(os.getcwd(), "foo", "newsfragments", "+.feature.rst")
        self.assertEqual(
            result.output,
            f"""Issue number: +
Fragment type (feature, bugfix, doc, removal, misc): feature
Created news fragment at {expected}
""",
        )
        with open(expected) as f:
            self.assertEqual(f.read(), "Edited content\n")

    @with_isolated_runner
    def test_sections(self, runner: CliRunner):
        """
        When creating a new fragment, the user can specify the section from the command
        line (and if none is provided, the default section will be used).

        The default section is either the section with a blank path, or else the first
        section defined in the configuration file.
        """
        setup_simple_project(
            extra_config="""
[[tool.towncrier.section]]
name = "Backend"
path = "backend"
[[tool.towncrier.section]]
name = "Frontend"
path = ""
"""
        )
        result = runner.invoke(_main, ["123.feature.rst"])
        self.assertFalse(result.exception, result.output)
        frag_path = Path("foo", "newsfragments")
        fragments = [f.name for f in frag_path.iterdir()]
        self.assertEqual(fragments, ["123.feature.rst"])

        result = runner.invoke(_main, ["123.feature.rst", "--section", "invalid"])
        self.assertTrue(result.exception, result.output)
        self.assertIn(
            "Invalid value for '--section': expected one of 'Backend', 'Frontend'",
            result.output,
        )

        result = runner.invoke(_main, ["123.feature.rst", "--section", "Backend"])
        self.assertFalse(result.exception, result.output)
        frag_path = Path("foo", "backend", "newsfragments")
        fragments = [f.name for f in frag_path.iterdir()]
        self.assertEqual(fragments, ["123.feature.rst"])

    @with_isolated_runner
    def test_sections_without_filename(self, runner: CliRunner):
        """
        When multiple sections exist when the interactive prompt is used, the user is
        prompted to select a section.
        """
        setup_simple_project(
            extra_config="""
[[tool.towncrier.section]]
name = "Backend"
path = ""

[[tool.towncrier.section]]
name = "Frontend"
path = "frontend"
"""
        )
        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "Edited content"
            result = runner.invoke(_main, input="2\n123\nfeature\n")
            self.assertFalse(result.exception, result.output)
            mock_edit.assert_called_once()
        expected = os.path.join(
            os.getcwd(), "foo", "frontend", "newsfragments", "123.feature.rst"
        )

        self.assertEqual(
            result.output,
            f"""\
Pick a section:
 1: Backend
 2: Frontend
Section (1, 2) [1]: 2
Issue number (`+` if none): 123
Fragment type (feature, bugfix, doc, removal, misc): feature
Created news fragment at {expected}
""",
        )

    @with_isolated_runner
    def test_sections_without_filename_with_section_option(self, runner: CliRunner):
        """
        When multiple sections exist and the section is provided via the command line,
        the user isn't prompted to select a section.
        """
        setup_simple_project(
            extra_config="""
[[tool.towncrier.section]]
name = "Backend"
path = ""

[[tool.towncrier.section]]
name = "Frontend"
path = "frontend"
"""
        )
        with mock.patch("click.edit") as mock_edit:
            mock_edit.return_value = "Edited content"
            result = runner.invoke(
                _main, ["--section", "Frontend"], input="123\nfeature\n"
            )
            self.assertFalse(result.exception, result.output)
            mock_edit.assert_called_once()
        expected = os.path.join(
            os.getcwd(), "foo", "frontend", "newsfragments", "123.feature.rst"
        )

        self.assertEqual(
            result.output,
            f"""\
Issue number (`+` if none): 123
Fragment type (feature, bugfix, doc, removal, misc): feature
Created news fragment at {expected}
""",
        )

    @with_isolated_runner
    def test_sections_all_with_paths(self, runner: CliRunner):
        """
        When all sections have paths, the first is the default.
        """
        setup_simple_project(
            extra_config="""
[[tool.towncrier.section]]
name = "Frontend"
path = "frontend"

[[tool.towncrier.section]]
name = "Backend"
path = "backend"
"""
        )
        result = runner.invoke(_main, ["123.feature.rst"])
        self.assertFalse(result.exception, result.output)
        frag_path = Path("foo", "frontend", "newsfragments")
        fragments = [f.name for f in frag_path.iterdir()]
        self.assertEqual(fragments, ["123.feature.rst"])

    @with_isolated_runner
    def test_without_filename_with_message(self, runner: CliRunner):
        """
        When no filename is provided, the user is prompted for one. If a message is
        provided, the editor isn't opened and the message is used.
        """
        setup_simple_project()

        with mock.patch("click.edit") as mock_edit:
            result = runner.invoke(_main, ["-c", "Fixed this"], input="123\nfeature\n")
            self.assertFalse(result.exception, result.output)
            mock_edit.assert_not_called()
        expected = os.path.join(os.getcwd(), "foo", "newsfragments", "123.feature.rst")
        self.assertEqual(
            result.output,
            f"""Issue number (`+` if none): 123
Fragment type (feature, bugfix, doc, removal, misc): feature
Created news fragment at {expected}
""",
        )
        with open(expected) as f:
            self.assertEqual(f.read(), "Fixed this\n")

    @with_isolated_runner
    def test_create_orphan_fragment(self, runner: CliRunner):
        """
        When a fragment starts with the only the orphan prefix (``+`` by default), the
        create CLI automatically extends the new file's base name to contain a random
        value to avoid commit collisions.
        """
        setup_simple_project()

        frag_path = Path("foo", "newsfragments")
        sub_frag_path = frag_path / "subsection"
        sub_frag_path.mkdir()

        result = runner.invoke(_main, ["+.feature"])
        self.assertEqual(0, result.exit_code)
        result = runner.invoke(
            _main, [str(Path("subsection", "+.feature"))], catch_exceptions=False
        )
        self.assertEqual(0, result.exit_code, result.output)

        fragments = [p for p in frag_path.rglob("*") if p.is_file()]
        self.assertEqual(2, len(fragments))
        change1, change2 = fragments

        self.assertEqual(change1.suffix, ".rst")
        self.assertTrue(change1.stem.startswith("+"))
        self.assertTrue(change1.stem.endswith(".feature"))
        # Length should be '+' character, 8 random hex characters, and ".feature".
        self.assertEqual(len(change1.stem), 1 + 8 + len(".feature"))

        self.assertEqual(change2.suffix, ".rst")
        self.assertTrue(change2.stem.startswith("+"))
        self.assertTrue(change2.stem.endswith(".feature"))
        self.assertEqual(change2.parent, sub_frag_path)
        # Length should be '+' character, 8 random hex characters, and ".feature".
        self.assertEqual(len(change2.stem), 1 + 8 + len(".feature"))

    @with_isolated_runner
    def test_create_orphan_fragment_custom_prefix(self, runner: CliRunner):
        """
        Check that the orphan prefix can be customized.
        """
        setup_simple_project(extra_config='orphan_prefix = "$$$"')

        frag_path = Path("foo", "newsfragments")

        result = runner.invoke(_main, ["$$$.feature"])
        self.assertEqual(0, result.exit_code, result.output)

        fragments = list(frag_path.rglob("*"))
        self.assertEqual(len(fragments), 1)
        change = fragments[0]
        self.assertTrue(change.stem.startswith("$$$"))
        # Length should be '$$$' characters, 8 random hex characters, and ".feature".
        self.assertEqual(len(change.stem), 3 + 8 + len(".feature"))
        # Check the remainder are all hex characters.
        self.assertTrue(
            all(c in string.hexdigits for c in change.stem[3 : -len(".feature")])
        )

    @with_isolated_runner
    def test_in_different_dir_with_nondefault_newsfragments_directory(self, runner):
        """
        When the `--dir` CLI argument is passed,
        it will create a new file in directory that is
        created by combining the `--dir` value
        with the `directory` option from the configuration
        file.
        """
        Path("pyproject.toml").write_text(
            # Important to customize `config.directory` because the default
            # already supports this scenario.
            "[tool.towncrier]\n"
            + 'directory = "changelog.d"\n'
        )
        Path("foo/foo").mkdir(parents=True)
        Path("foo/foo/__init__.py").write_text("")

        result = runner.invoke(
            _main,
            (
                "--config",
                "pyproject.toml",
                "--dir",
                "foo",
                "--content",
                "Adds levitation.",
                "123.feature",
            ),
        )

        self.assertEqual(0, result.exit_code)
        self.assertTrue(Path("foo/changelog.d/123.feature.rst").exists())



from __future__ import annotations

import os

import nox


nox.options.sessions = ["pre_commit", "docs", "typecheck", "tests"]
nox.options.reuse_existing_virtualenvs = True
nox.options.error_on_external_run = True


@nox.session
def pre_commit(session: nox.Session) -> None:
    session.install("pre-commit")

    session.run("pre-commit", "run", "--all-files", "--show-diff-on-failure")


# Keep list in-sync with ci.yml/test-linux & pyproject.toml
@nox.session(python=["pypy3.8", "3.8", "3.9", "3.10", "3.11", "3.12"])
def tests(session: nox.Session) -> None:
    session.env["PYTHONWARNDEFAULTENCODING"] = "1"
    session.install("Twisted", "coverage[toml]")
    posargs = list(session.posargs)

    try:
        # Allow `--use-wheel path/to/wheel.whl` to be passed.
        i = session.posargs.index("--use-wheel")
        session.install(session.posargs[i + 1])
        del posargs[i : i + 2]
    except ValueError:
        session.install(".")

    if not posargs:
        posargs = ["towncrier"]

    session.run("coverage", "run", "--module", "twisted.trial", *posargs)

    if os.environ.get("CI") != "true":
        session.notify("coverage_report")


@nox.session
def coverage_report(session: nox.Session) -> None:
    session.install("coverage[toml]")

    session.run("coverage", "combine")
    session.run("coverage", "report")


@nox.session
def check_newsfragment(session: nox.Session) -> None:
    session.install(".")
    session.run("python", "-m", "towncrier.check", "--compare-with", "origin/trunk")


@nox.session
def draft_newsfragment(session: nox.Session) -> None:
    session.install(".")
    session.run("python", "-m", "towncrier.build", "--draft")


@nox.session
def typecheck(session: nox.Session) -> None:
    # Click 8.1.4 is bad type hints -- lets not complicate packaging and only
    # pin here.
    session.install(".", "mypy", "click!=8.1.4")
    session.run("mypy", "src")


@nox.session
def docs(session: nox.Session) -> None:
    session.install(".[dev]")

    session.run(
        # fmt: off
        "python", "-m", "sphinx",
        "-T", "-E",
        "-W", "--keep-going",
        "-b", "html",
        "-d", "docs/_build/doctrees",
        "-D", "language=en",
        "docs",
        "docs/_build/html",
        # fmt: on
    )


@nox.session
def build(session: nox.Session) -> None:
    session.install("build", "twine")

    # If no argument is passed, build builds an sdist and then a wheel from
    # that sdist.
    session.run("python", "-m", "build")

    session.run("twine", "check", "--strict", "dist/*")



from __future__ import annotations

import sys
import textwrap

from functools import wraps
from pathlib import Path
from subprocess import call
from typing import Any, Callable

from click.testing import CliRunner


if sys.version_info < (3, 9):
    import importlib_resources as resources
else:
    from importlib import resources


def read(filename: str | Path) -> str:
    return Path(filename).read_text()


def write(path: str | Path, contents: str, dedent: bool = False) -> None:
    """
    Create a file with given contents including any missing parent directories
    """
    p = Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    if dedent:
        contents = textwrap.dedent(contents)
    p.write_text(contents)


def read_pkg_resource(path: str) -> str:
    """
    Read *path* from the towncrier package.
    """
    return (resources.files("towncrier") / path).read_text("utf-8")


def with_isolated_runner(fn: Callable[..., Any]) -> Callable[..., Any]:
    """
    Run *fn* within an isolated filesystem and add the kwarg *runner* to its
    arguments.
    """

    @wraps(fn)
    def test(*args: Any, **kw: Any) -> Any:
        runner = CliRunner()
        with runner.isolated_filesystem():
            return fn(*args, runner=runner, **kw)

    return test


def setup_simple_project(
    *,
    config: str | None = None,
    extra_config: str = "",
    pyproject_path: str = "pyproject.toml",
    mkdir_newsfragments: bool = True,
) -> None:
    if config is None:
        config = "[tool.towncrier]\n" 'package = "foo"\n' + extra_config
    else:
        config = textwrap.dedent(config)
    Path(pyproject_path).write_text(config)
    Path("foo").mkdir()
    Path("foo/__init__.py").write_text('__version__ = "1.2.3"\n')

    if mkdir_newsfragments:
        Path("foo/newsfragments").mkdir()


def with_project(
    *,
    config: str | None = None,
    pyproject_path: str = "pyproject.toml",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to run a test with an isolated directory containing a simple
    project.

    The files are not managed by git.

    `config` is the content of the config file.
    It will be automatically dedented.

    `pyproject_path` is the path where to store the config file.
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(fn)
        def test(*args: Any, **kw: Any) -> Any:
            runner = CliRunner()
            with runner.isolated_filesystem():
                setup_simple_project(
                    config=config,
                    pyproject_path=pyproject_path,
                )

                return fn(*args, runner=runner, **kw)

        return test

    return decorator


def with_git_project(
    *,
    config: str | None = None,
    pyproject_path: str = "pyproject.toml",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to run a test with an isolated directory containing a simple
    project checked into git.
    Use `config` to tweak the content of the config file.
    Use `pyproject_path` to tweak the location of the config file.
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        def _commit() -> None:
            call(["git", "add", "."])
            call(["git", "commit", "-m", "Second Commit"])

        @wraps(fn)
        def test(*args: Any, **kw: Any) -> Any:
            runner = CliRunner()
            with runner.isolated_filesystem():
                setup_simple_project(
                    config=config,
                    pyproject_path=pyproject_path,
                )

                call(["git", "init"])
                call(["git", "config", "user.name", "user"])
                call(["git", "config", "user.email", "user@example.com"])
                call(["git", "config", "commit.gpgSign", "false"])
                call(["git", "add", "."])
                call(["git", "commit", "-m", "Initial Commit"])

                return fn(*args, runner=runner, commit=_commit, **kw)

        return test

    return decorator



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

from __future__ import annotations

import atexit
import dataclasses
import os
import re
import sys

from contextlib import ExitStack
from pathlib import Path
from typing import Any, Literal, Mapping, Sequence

from click import ClickException

from .._settings import fragment_types as ft


if sys.version_info < (3, 10):
    import importlib_resources as resources
else:
    from importlib import resources


if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib


re_resource_template = re.compile(r"[-\w.]+:[-\w.]+$")


@dataclasses.dataclass
class Config:
    sections: Mapping[str, str]
    types: Mapping[str, Mapping[str, Any]]
    template: str | tuple[str, str]
    start_string: str
    package: str = ""
    package_dir: str = "."
    single_file: bool = True
    filename: str = "NEWS.rst"
    directory: str | None = None
    version: str | None = None
    name: str = ""
    title_format: str | Literal[False] = ""
    issue_format: str | None = None
    underlines: Sequence[str] = ("=", "-", "~")
    wrap: bool = False
    all_bullets: bool = True
    orphan_prefix: str = "+"
    create_eof_newline: bool = True
    create_add_extension: bool = True
    ignore: list[str] | None = None


class ConfigError(ClickException):
    def __init__(self, *args: str, **kwargs: str):
        self.failing_option = kwargs.get("failing_option")
        super().__init__(*args)


def load_config_from_options(
    directory: str | None, config_path: str | None
) -> tuple[str, Config]:
    """
    Load the configuration from a given directory or specific configuration file.

    Unless an explicit configuration file is given, traverse back from the given
    directory looking for a configuration file.

    Returns a tuple of the base directory and the parsed Config instance.
    """
    if config_path is None:
        return traverse_for_config(directory)

    config_path = os.path.abspath(config_path)

    # When a directory is provided (in addition to the config file), use it as the base
    # directory. Otherwise use the directory containing the config file.
    if directory is not None:
        base_directory = os.path.abspath(directory)
    else:
        base_directory = os.path.dirname(config_path)

    if not os.path.isfile(config_path):
        raise ConfigError(f"Configuration file '{config_path}' not found.")
    config = load_config_from_file(base_directory, config_path)

    return base_directory, config


def traverse_for_config(path: str | None) -> tuple[str, Config]:
    """
    Search for a configuration file in the current directory and all parent directories.

    Returns the directory containing the configuration file and the parsed configuration.
    """
    start_directory = directory = os.path.abspath(path or os.getcwd())
    while True:
        config = load_config(directory)
        if config is not None:
            return directory, config

        parent = os.path.dirname(directory)
        if parent == directory:
            raise ConfigError(
                f"No configuration file found.\nLooked back from: {start_directory}"
            )
        directory = parent


def load_config(directory: str) -> Config | None:
    towncrier_toml = os.path.join(directory, "towncrier.toml")
    pyproject_toml = os.path.join(directory, "pyproject.toml")

    if os.path.exists(towncrier_toml):
        config_file = towncrier_toml
    elif os.path.exists(pyproject_toml):
        config_file = pyproject_toml
    else:
        return None

    return load_config_from_file(directory, config_file)


def load_config_from_file(directory: str, config_file: str) -> Config:
    with open(config_file, "rb") as conffile:
        config = tomllib.load(conffile)

    return parse_toml(directory, config)


# Clean up possible temporary files on exit.
_file_manager = ExitStack()
atexit.register(_file_manager.close)


def parse_toml(base_path: str, config: Mapping[str, Any]) -> Config:
    if "towncrier" not in (config.get("tool") or {}):
        raise ConfigError("No [tool.towncrier] section.", failing_option="all")

    config = config["tool"]["towncrier"]
    parsed_data = {}

    # Check for misspelt options.
    for typo, correct in [
        ("singlefile", "single_file"),
    ]:
        if config.get(typo):
            raise ConfigError(
                f"`{typo}` is not a valid option. Did you mean `{correct}`?",
                failing_option=typo,
            )

    # Process options.
    for field in dataclasses.fields(Config):
        if field.name in ("sections", "types", "template"):
            # Skip these options, they are processed later.
            continue
        if field.name in config:
            #  Interestingly, the __future__ annotation turns the type into a string.
            if field.type in ("bool", bool):
                if not isinstance(config[field.name], bool):
                    raise ConfigError(
                        f"`{field.name}` option must be boolean: false or true.",
                        failing_option=field.name,
                    )
            parsed_data[field.name] = config[field.name]

    # Process 'section'.
    sections = {}
    if "section" in config:
        for x in config["section"]:
            sections[x.get("name", "")] = x["path"]
    else:
        sections[""] = ""
    parsed_data["sections"] = sections

    # Process 'types'.
    fragment_types_loader = ft.BaseFragmentTypesLoader.factory(config)
    parsed_data["types"] = fragment_types_loader.load()

    # Process 'template'.
    markdown_file = Path(config.get("filename", "")).suffix == ".md"
    template = config.get("template", "towncrier:default")
    if re_resource_template.match(template):
        package, resource = template.split(":", 1)
        if not Path(resource).suffix:
            resource += ".md" if markdown_file else ".rst"

        if not _pkg_file_exists(package, resource):
            if _pkg_file_exists(package + ".templates", resource):
                package += ".templates"
            else:
                raise ConfigError(
                    f"'{package}' does not have a template named '{resource}'.",
                    failing_option="template",
                )
        template = (package, resource)
    else:
        template = os.path.join(base_path, template)
        if not os.path.isfile(template):
            raise ConfigError(
                f"The template file '{template}' does not exist.",
                failing_option="template",
            )

    parsed_data["template"] = template

    # Process 'start_string'.

    start_string = config.get("start_string", "")
    if not start_string:
        start_string_template = "<!-- {} -->\n" if markdown_file else ".. {}\n"
        start_string = start_string_template.format("towncrier release notes start")
    parsed_data["start_string"] = start_string

    # Return the parsed config.
    return Config(**parsed_data)


def _pkg_file_exists(pkg: str, file: str) -> bool:
    """
    Check whether *file* exists within *pkg*.
    """
    return resources.files(pkg).joinpath(file).is_file()
